import pandas as pd
import pytest
from swapanything_backend import _base


def test_incomplete_base_backend():
    AVAIL_COL = "avail"
    SUBJ_COL = "subj"
    EXCL_SUBJ_COLS = ["s1", "s2"]

    class TestBackend(_base.BackendBase):
        def __init__(self) -> None:
            self.subject_features = ["a", "b", "c"]
            self.availability_subject_column = SUBJ_COL
            self.availabilities_column = AVAIL_COL
            self.exclusions_subject_columns = EXCL_SUBJ_COLS

    with pytest.raises(TypeError, match=r"Can't instantiate abstract class"):
        TestBackend()


def test_base_backend():
    AVAIL_COL = "avail"
    SUBJ_COL = "subj"
    EXCL_SUBJ_COLS = ["s1", "s2"]

    class TestBackend(_base.BackendBase):
        def __init__(self) -> None:
            self.subject_features = ["a", "b", "c"]
            self.availability_subject_column = SUBJ_COL
            self.availabilities_column = AVAIL_COL
            self.exclusions_subject_columns = EXCL_SUBJ_COLS

        def get_subjects(self) -> None:
            return

        def get_availabilities(self) -> pd.DataFrame:
            return

        def get_exclusions(self) -> pd.DataFrame:
            return

    TestBackend()



from abc import ABC, abstractmethod
from typing import Annotated, Iterable, TypeVar

import pandas as pd


class BackendError(Exception):
    pass


class BackendBase(ABC):
    subject_features: Iterable[str]
    availability_subject_column: str
    availabilities_column: str
    exclusions_subject_columns: Annotated[Iterable[str], 2]

    @abstractmethod
    def get_subjects(self, *args, **kwargs) -> pd.DataFrame:  # pragma: no cover
        raise NotImplementedError()

    @abstractmethod
    def get_availabilities(self, *args, **kwargs) -> pd.DataFrame:  # pragma: no cover
        raise NotImplementedError()

    @abstractmethod
    def get_exclusions(self, *args, **kwargs) -> pd.DataFrame:  # pragma: no cover
        raise NotImplementedError()


BackendType = TypeVar("BackendType", bound=BackendBase)



from itertools import combinations
from typing import Annotated, Iterable, Optional, Tuple, Union

import pandas as pd


class BackendError(Exception):
    pass


def get_matching_subjects_by_slot(
    availabilities: pd.DataFrame,
    slot_col: str,
    subject_col: str,
) -> pd.DataFrame:
    matching_subjects_by_slot = (
        availabilities.groupby(slot_col)
        .agg({subject_col: lambda x: tuple(x)})
        .reset_index(col_level=1)
    )
    matching_subjects_by_slot = matching_subjects_by_slot[
        matching_subjects_by_slot[subject_col].apply(len) > 1
    ]
    return matching_subjects_by_slot


def get_matches_from_slots(
    matching_subjects_by_slot: pd.DataFrame,
    slot_col: str,
    subject_col: str,
) -> pd.DataFrame:
    matches = (
        matching_subjects_by_slot.set_index(slot_col)[subject_col]
        .apply(lambda a: list(combinations(a, 2)))
        .reset_index()
        .explode(subject_col)
    )
    matches[subject_col] = matches[subject_col].apply(sorted).apply(tuple)
    return matches


def apply_exclusions(
    matches: pd.DataFrame,
    exclusions: pd.DataFrame,
    subject_col: str,
    exclusions_subject_columns: Annotated[Iterable[str], 2],
) -> pd.DataFrame:
    exclude_matchings = exclusions[exclusions_subject_columns]
    exclude_matchings = [tuple(sorted(x)) for x in exclude_matchings.values]
    exluded_mask = matches[subject_col].isin(exclude_matchings)
    matches = matches.loc[~exluded_mask]
    return matches


def get_all_matches(
    availabilities: pd.DataFrame,
    subject_col: str,
    slot_col: str,
    subjects_new_col_name: Optional[str] = None,
    slots_new_col_name: Optional[str] = None,
    exclusions: Union[pd.DataFrame, None] = None,
    exclusions_subject_columns: Tuple[str, str] = None,
    return_matching_subjects_by_slot: bool = False,
    availabilities_by_slot: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
    if not availabilities_by_slot:
        matching_subjects_by_slot = get_matching_subjects_by_slot(
            availabilities,
            slot_col=slot_col,
            subject_col=subject_col,
        )
    else:
        matching_subjects_by_slot = availabilities

    matches = get_matches_from_slots(
        matching_subjects_by_slot,
        slot_col=slot_col,
        subject_col=subject_col,
    )

    if isinstance(exclusions, pd.DataFrame):
        matches = apply_exclusions(
            matches,
            exclusions=exclusions,
            subject_col=subject_col,
            exclusions_subject_columns=exclusions_subject_columns,
        )

    # go from:
    #  [{"avail": "A", "subj": (1, 2)},
    #   {"avail": "B", "subj": (1, 2)}]
    # to:
    #  [{"subj": (1, 2), "avail": ("A", "B")}]
    matches = matches.sort_values(
        # Sort to guarantee idempotency downstream
        [slot_col]
    )
    matches = matches.groupby(subject_col)[[slot_col]].agg(tuple).reset_index()

    if slots_new_col_name:
        matches = matches.rename(columns={slot_col: slots_new_col_name})

    if subjects_new_col_name:
        matches = matches.rename(columns={subject_col: subjects_new_col_name})

    if return_matching_subjects_by_slot:
        return matches, matching_subjects_by_slot
    else:
        return matches



import pandas as pd
from swapanything import prep


def test__get_matching_subjects_by_slot():
    SLOT_COL = "avail"
    SUBJ_COL = "subj"
    df = pd.DataFrame(
        [
            ["sub1", "A"],
            ["sub2", "B"],
            ["sub5", "A"],
            ["sub3", "A"],
            ["sub4", "C"],
        ],
        columns=[SUBJ_COL, SLOT_COL],
    )

    expected_result = pd.DataFrame(
        [
            ["A", ("sub1", "sub5", "sub3")],
        ],
        columns=[SLOT_COL, SUBJ_COL],
    )

    result = prep.get_matching_subjects_by_slot(
        availabilities=df,
        subject_col=SUBJ_COL,
        slot_col=SLOT_COL,
    )
    assert isinstance(result, pd.DataFrame)
    assert result.equals(expected_result)


def test__get_matches_from_slots():
    SLOT_COL = "avail"
    SUBJ_COL = "subj"
    slots = pd.DataFrame(
        [
            ["A", ("sub1", "sub5", "sub3")],
        ],
        columns=[SLOT_COL, SUBJ_COL],
    )

    expected_result = pd.DataFrame(
        [
            ["A", ("sub1", "sub5")],
            ["A", ("sub1", "sub3")],
            ["A", ("sub3", "sub5")],
        ],
        columns=[SLOT_COL, SUBJ_COL],
        index=[0, 0, 0],
    )

    result = prep.get_matches_from_slots(slots, slot_col=SLOT_COL, subject_col=SUBJ_COL)
    assert isinstance(result, pd.DataFrame)
    assert result.equals(expected_result)


def test__apply_exclusions():
    SUBJ_COL = "subj"
    EXCL_SUBJ_COLS = ["s1", "s2"]

    matches = pd.DataFrame(
        [
            ["A", ("sub1", "sub5")],
            ["A", ("sub1", "sub3")],
            ["A", ("sub3", "sub5")],
        ],
        columns=["avail", SUBJ_COL],
        index=[0, 0, 0],
    )
    exclusions = pd.DataFrame(
        [
            ["e1", "sub1", "sub3"],
        ],
        columns=["excl", EXCL_SUBJ_COLS[0], EXCL_SUBJ_COLS[1]],
    )
    expected_result = matches.loc[[True, False, True]]

    result = prep.apply_exclusions(
        matches,
        exclusions,
        subject_col=SUBJ_COL,
        exclusions_subject_columns=EXCL_SUBJ_COLS,
    )
    assert isinstance(result, pd.DataFrame)
    assert result.equals(expected_result)


def test_get_all_matches():
    SLOT_COL = "avail"
    SUBJ_COL = "subj"
    EXCL_SUBJ_COLS = ["s1", "s2"]
    availabilities = pd.DataFrame(
        [
            ["sub3", "A"],
            ["sub1", "A"],
            ["sub2", "B"],
            ["sub5", "A"],
            ["sub4", "C"],
        ],
        columns=[SUBJ_COL, SLOT_COL],
    )

    exclusions = pd.DataFrame(
        [
            ["e1", "sub1", "sub3"],
        ],
        columns=["excl", EXCL_SUBJ_COLS[0], EXCL_SUBJ_COLS[1]],
    )

    expected_result = pd.DataFrame(
        [
            [("sub1", "sub5"), ("A",)],
            [("sub3", "sub5"), ("A",)],
        ],
        columns=[SUBJ_COL, "avail"],
    )

    result = prep.get_all_matches(
        availabilities=availabilities,
        subject_col=SUBJ_COL,
        slot_col=SLOT_COL,
        exclusions=exclusions,
        exclusions_subject_columns=EXCL_SUBJ_COLS,
    )
    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {SUBJ_COL, SLOT_COL}
    assert result.equals(expected_result)

    availabilities_by_slot = prep.get_matching_subjects_by_slot(
        availabilities=availabilities,
        subject_col=SUBJ_COL,
        slot_col=SLOT_COL,
    )
    result2 = prep.get_all_matches(
        availabilities=availabilities_by_slot,
        subject_col=SUBJ_COL,
        slot_col=SLOT_COL,
        exclusions=exclusions,
        exclusions_subject_columns=EXCL_SUBJ_COLS,
        availabilities_by_slot=True,
    )
    assert result2.equals(result)

    result3, matching_subjects_by_slot = prep.get_all_matches(
        availabilities=availabilities,
        subject_col=SUBJ_COL,
        slot_col=SLOT_COL,
        exclusions=exclusions,
        exclusions_subject_columns=EXCL_SUBJ_COLS,
        return_matching_subjects_by_slot=True,
        slots_new_col_name="slots_new_col_name",
        subjects_new_col_name="subjects_new_col_name",
    )
    assert isinstance(result3, pd.DataFrame)
    assert result3.to_dict(orient="records") == [
        {"subjects_new_col_name": ("sub1", "sub5"), "slots_new_col_name": ("A",)},
        {"subjects_new_col_name": ("sub3", "sub5"), "slots_new_col_name": ("A",)},
    ]
    assert matching_subjects_by_slot.to_dict(orient="records") == [
        {"avail": "A", "subj": ("sub3", "sub1", "sub5")}
    ]



from collections.abc import Iterator
from typing import Annotated, Iterable, List

import pandas as pd
import pyairtable as airtable
from pydantic import BaseModel, Field, PastDatetime, SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict

# from swapanything import models as m
from ._base import BackendBase


class AirtableRecord(BaseModel):
    record_id: str = Field(alias="id")
    created_timestamp: PastDatetime = Field(alias="createdTime")
    record_fields: dict = Field(alias="fields")


class AirtableResponse(BaseModel):
    records: List[AirtableRecord]

    def iter_records(self) -> Iterator[dict]:
        for record in self.records:
            yield record.record_fields

    def to_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame(self.iter_records())
        return df


class AirTableBackend(BaseSettings, BackendBase):
    # Manually adding again to help pylance and other
    # linters not to get crazy
    subject_features: List[str]
    availability_subject_column: str
    availabilities_column: str
    exclusions_subject_columns: Annotated[Iterable[str], 2]

    subjects_table_name: str
    availabilities_table_name: str
    exclusions_table_name: str
    client_id: SecretStr = Field(..., validation_alias="AIRTABLE_BASE_ID")
    client_secret: SecretStr = Field(..., validation_alias="AIRTABLE_API_KEY")

    model_config = SettingsConfigDict(case_sensitive=False, env_prefix="SWPAT_")

    def _get_table(
        self,
        table_name: str,
        **options,
    ) -> List[dict]:
        table = airtable.Table(  # type: ignore
            api_key=self.client_secret.get_secret_value(),
            base_id=self.client_id.get_secret_value(),
            table_name=table_name,
        )

        all = table.all(**options)
        return all

    def get_subjects(self, formula: str | None = None) -> pd.DataFrame:
        table = self._get_table(
            table_name=self.subjects_table_name,
            fields=["recID", *self.subject_features],
            formula=formula,
        )
        table = AirtableResponse.model_validate({"records": table})
        table = (
            table.to_pandas()
            .rename(columns={"recID": "subject_id"})
            .set_index("subject_id")
        )
        return table

    def get_exclusions(self, formula: str | None = None) -> pd.DataFrame:
        table = self._get_table(
            table_name=self.exclusions_table_name,
            fields=["recID", *self.exclusions_subject_columns],
            formula=formula,
        )
        table = AirtableResponse.model_validate({"records": table})
        table = (
            table.to_pandas()
            .rename(columns={"recID": "exclusion_id"})
            .set_index("exclusion_id")
        )
        return table

    def get_availabilities(self, formula: str | None = None) -> pd.DataFrame:
        table = self._get_table(
            table_name=self.availabilities_table_name,
            fields=[
                "recID",
                self.availability_subject_column,
                self.availabilities_column,
            ],
            formula=formula,
        )
        table = AirtableResponse.model_validate({"records": table})
        table = (
            table.to_pandas()
            .rename(
                columns={
                    "recID": "availability_id",
                    self.availability_subject_column: "subject_id",
                    self.availabilities_column: "availabilities",
                }
            )
            .set_index("availability_id")
        )
        return table



import re
from typing import List
from unittest.mock import MagicMock

import pandas as pd
import pytest
from swapanything_backend import airtable as be


@pytest.fixture
def BASE_ID() -> str:
    return "baseId"


@pytest.fixture
def API_KEY() -> str:
    return "apiKey"


@pytest.fixture
def mock_airtable_env_vars(
    monkeypatch: pytest.MonkeyPatch, BASE_ID: str, API_KEY: str
) -> None:
    monkeypatch.setenv("AIRTABLE_BASE_ID", BASE_ID)
    monkeypatch.setenv("AIRTABLE_API_KEY", API_KEY)


@pytest.fixture
def mock_airtable(
    monkeypatch: pytest.MonkeyPatch,
    CLEAN_SUBJECTS_RECORDS: List[dict],
    CLEAN_AVAILABILITIES_RECORDS: List[dict],
    CLEAN_EXCLUSIONS_RECORDS: List[dict],
    SUBJECTS_TABLE_NAME: str,
    AVAILABILITIES_TABLE_NAME: str,
    EXCLUSIONS_TABLE_NAME: str,
    mock_airtable_env_vars: None,
) -> MagicMock:
    """Make Airtable SDK return our mock subjects records."""
    # custom class to be the mock return value of requests.get()
    mock_airtable = MagicMock(be.airtable.Table)  # type: ignore
    mock_airtable.return_value = mock_airtable

    def get_records(*args, **kwargs) -> List[dict]:
        table_name = mock_airtable.call_args[1]["table_name"]
        if table_name == SUBJECTS_TABLE_NAME:
            return CLEAN_SUBJECTS_RECORDS
        elif table_name == AVAILABILITIES_TABLE_NAME:
            return CLEAN_AVAILABILITIES_RECORDS
        elif table_name == EXCLUSIONS_TABLE_NAME:
            return CLEAN_EXCLUSIONS_RECORDS
        else:
            raise NotImplementedError("No mock records for '%s'" % table_name)

    mock_airtable.all.side_effect = get_records

    monkeypatch.setattr("pyairtable.Table", mock_airtable)
    return mock_airtable


@pytest.fixture
def airtable_backend(
    SUBJECT_FEATURES: List[str],
    AVAILABILITY_SUBJECT_COLUMN: str,
    AVAILABILITIES_COLUMN: str,
    SUBJECTS_TABLE_NAME: str,
    AVAILABILITIES_TABLE_NAME: str,
    EXCLUSIONS_TABLE_NAME: str,
    EXCLUSOIONS_SUBJECT_COLUMNS: List[str],
    mock_airtable: None,
) -> be.AirTableBackend:
    backend = be.AirTableBackend(  # type: ignore
        subject_features=SUBJECT_FEATURES,
        availability_subject_column=AVAILABILITY_SUBJECT_COLUMN,
        availabilities_column=AVAILABILITIES_COLUMN,
        subjects_table_name=SUBJECTS_TABLE_NAME,
        availabilities_table_name=AVAILABILITIES_TABLE_NAME,
        exclusions_table_name=EXCLUSIONS_TABLE_NAME,
        exclusions_subject_columns=EXCLUSOIONS_SUBJECT_COLUMNS,
    )
    return backend


def test_backend_init(mock_airtable: MagicMock, BASE_ID: str, API_KEY: str):
    data = be.AirTableBackend(  # type: ignore
        subject_features=["a", "b"],
        availability_subject_column="availability_subject_column",
        availabilities_column="availabilities_column",
        subjects_table_name="subjects_table",
        availabilities_table_name="availabilities_table",
        exclusions_table_name="exclustions_table",
        exclusions_subject_columns=[
            "exclusions_subject_column1",
            "exclusions_subject_column2",
        ],
    )
    assert re.match(r"^\*{2,}$", str(data.client_id))  # all is more than 2 '*'
    assert re.match(r"^\*{2,}$", str(data.client_secret))
    assert data.client_id.get_secret_value() == BASE_ID
    assert data.client_secret.get_secret_value() == API_KEY


def test__get_table(
    mock_airtable: MagicMock,
    BASE_ID: str,
    API_KEY: str,
    SUBJECTS_TABLE_NAME: str,
    CLEAN_SUBJECTS_RECORDS: List[dict],
    SUBJECT_FEATURES: str,
) -> None:
    """Test the general method to retrieve a table using pyairtable SDK."""
    # Mock properties of initialized AirTableBackend
    airtable_backend_mock = MagicMock()
    airtable_backend_mock.client_secret.get_secret_value.return_value = API_KEY
    airtable_backend_mock.client_id.get_secret_value.return_value = BASE_ID

    # execute method only, using 'airtable_backend_mock' as 'self'
    table = be.AirTableBackend._get_table(
        airtable_backend_mock, SUBJECTS_TABLE_NAME, fields=SUBJECT_FEATURES
    )

    # Check that everything has been called properly
    mock_airtable.assert_called_once_with(
        api_key=API_KEY, base_id=BASE_ID, table_name=SUBJECTS_TABLE_NAME
    )
    mock_airtable.all.assert_called_once_with(fields=SUBJECT_FEATURES)

    # Check that results correspond
    assert table == CLEAN_SUBJECTS_RECORDS


def test_get_subjects(
    airtable_backend: be.AirTableBackend,
    mock_airtable: MagicMock,
    SUBJECTS_TABLE_NAME: str,
    SUBJECT_FEATURES: List[str],
    SUBJECTS_DF: pd.DataFrame,
    BASE_ID: str,
    API_KEY: str,
):
    df = airtable_backend.get_subjects(formula=None)
    mock_airtable.assert_called_once_with(
        api_key=API_KEY, base_id=BASE_ID, table_name=SUBJECTS_TABLE_NAME
    )
    mock_airtable.all.assert_called_once_with(
        fields=["recID"] + SUBJECT_FEATURES, formula=None
    )
    assert set(df.columns) == set(SUBJECT_FEATURES)
    assert df.equals(SUBJECTS_DF)


def test_get_availabilities(
    airtable_backend: be.AirTableBackend,
    mock_airtable: MagicMock,
    AVAILABILITIES_TABLE_NAME: str,
    AVAILABILITIES_DF: pd.DataFrame,
    AVAILABILITY_SUBJECT_COLUMN: str,
    AVAILABILITIES_COLUMN: str,
    BASE_ID: str,
    API_KEY: str,
):
    AVAIL_FEATURES = ["recID", AVAILABILITY_SUBJECT_COLUMN, AVAILABILITIES_COLUMN]
    df = airtable_backend.get_availabilities(formula=None)

    mock_airtable.assert_called_once_with(
        api_key=API_KEY, base_id=BASE_ID, table_name=AVAILABILITIES_TABLE_NAME
    )
    mock_airtable.all.assert_called_once_with(fields=AVAIL_FEATURES, formula=None)
    assert set(df.columns) == {"subject_id", "availabilities"}
    assert df.index.name == "availability_id"
    assert df.equals(AVAILABILITIES_DF)


def test_get_exclustions(
    airtable_backend: be.AirTableBackend,
    mock_airtable: MagicMock,
    EXCLUSIONS_TABLE_NAME: str,
    EXCLUSOIONS_SUBJECT_COLUMNS: List[str],
    EXCLUSIONS_DF: pd.DataFrame,
    BASE_ID: str,
    API_KEY: str,
):
    df = airtable_backend.get_exclusions(formula=None)
    mock_airtable.assert_called_once_with(
        api_key=API_KEY, base_id=BASE_ID, table_name=EXCLUSIONS_TABLE_NAME
    )
    mock_airtable.all.assert_called_once_with(
        fields=["recID"] + EXCLUSOIONS_SUBJECT_COLUMNS, formula=None
    )
    assert set(df.columns) == set(EXCLUSOIONS_SUBJECT_COLUMNS)
    assert df.equals(EXCLUSIONS_DF)



import json
from copy import deepcopy
from pathlib import Path
from typing import Annotated, List

import pandas as pd
import pytest
from typing_extensions import TypedDict

__here__ = Path(__file__).parent
__data__ = __here__.parent / "data"


class AirtableRecord(TypedDict):
    id: str
    createdTime: str
    fields: dict


@pytest.fixture
def SUBJECTS_TABLE_NAME() -> str:
    return "subjects"


@pytest.fixture
def AVAILABILITIES_TABLE_NAME() -> str:
    return "availabilities"


@pytest.fixture
def EXCLUSIONS_TABLE_NAME() -> str:
    return "exclusions"


@pytest.fixture
def SUBJECT_FEATURES() -> List[str]:
    subject_features = [
        "Growth Stage",
        "Goals",
        "Ruolo",
        "Customers",
        "Interessi",
        "Lingue Parlate",
    ]
    return subject_features


@pytest.fixture
def AVAILABILITY_SUBJECT_COLUMN() -> str:
    return "subj"


@pytest.fixture
def AVAILABILITIES_COLUMN() -> str:
    return "avail"


@pytest.fixture
def EXCLUSOIONS_SUBJECT_COLUMNS() -> Annotated[List[str], 2]:
    return ["subj1", "subj2"]


@pytest.fixture
def SUBJECTS_RECORDS() -> List[AirtableRecord]:
    return json.loads((__data__ / "subjects_records.json").read_bytes())


@pytest.fixture
def AVAILABILITIES_RECORDS() -> List[AirtableRecord]:
    return json.loads((__data__ / "availabilities_records.json").read_bytes())


@pytest.fixture
def EXCLUSIONS_RECORDS() -> List[AirtableRecord]:
    return json.loads((__data__ / "exclusions_records.json").read_bytes())


def _cleanup_record(
    r: AirtableRecord,
    cols: List[str],
) -> AirtableRecord:
    r = deepcopy(r)
    r["fields"] = {k: v for k, v in r["fields"].items() if k in cols}
    return r


@pytest.fixture
def CLEAN_SUBJECTS_RECORDS(
    SUBJECTS_RECORDS: List[AirtableRecord], SUBJECT_FEATURES: List[str]
) -> List[AirtableRecord]:
    rec = [_cleanup_record(r, ["recID"] + SUBJECT_FEATURES) for r in SUBJECTS_RECORDS]
    return rec


@pytest.fixture
def CLEAN_EXCLUSIONS_RECORDS(
    EXCLUSIONS_RECORDS: List[AirtableRecord], EXCLUSOIONS_SUBJECT_COLUMNS: List[str]
) -> List[AirtableRecord]:
    rec = [
        _cleanup_record(r, ["recID"] + EXCLUSOIONS_SUBJECT_COLUMNS)
        for r in EXCLUSIONS_RECORDS
    ]
    return rec


@pytest.fixture
def CLEAN_AVAILABILITIES_RECORDS(
    AVAILABILITIES_RECORDS: List[AirtableRecord],
    AVAILABILITIES_COLUMN: str,
    AVAILABILITY_SUBJECT_COLUMN: str,
) -> List[AirtableRecord]:
    rec = [
        _cleanup_record(
            r, ["recID", AVAILABILITIES_COLUMN, AVAILABILITY_SUBJECT_COLUMN]
        )
        for r in AVAILABILITIES_RECORDS
    ]
    return rec


@pytest.fixture
def SUBJECTS_DF(CLEAN_SUBJECTS_RECORDS: List[AirtableRecord]) -> pd.DataFrame:
    records = (
        pd.DataFrame(r["fields"] for r in CLEAN_SUBJECTS_RECORDS)
        .rename(columns={"recID": "subject_id"})
        .set_index("subject_id")
    )
    return records


@pytest.fixture
def AVAILABILITIES_DF(
    CLEAN_AVAILABILITIES_RECORDS: List[AirtableRecord],
    AVAILABILITY_SUBJECT_COLUMN: str,
    AVAILABILITIES_COLUMN: str,
) -> pd.DataFrame:
    records = (
        pd.DataFrame(r["fields"] for r in CLEAN_AVAILABILITIES_RECORDS)
        .rename(
            columns={
                "recID": "availability_id",
                AVAILABILITY_SUBJECT_COLUMN: "subject_id",
                AVAILABILITIES_COLUMN: "availabilities",
            }
        )
        .set_index("availability_id")
    )
    return records


@pytest.fixture
def EXCLUSIONS_DF(CLEAN_EXCLUSIONS_RECORDS: List[AirtableRecord]) -> pd.DataFrame:
    records = (
        pd.DataFrame(r["fields"] for r in CLEAN_EXCLUSIONS_RECORDS)
        .rename(columns={"recID": "exclusion_id"})
        .set_index("exclusion_id")
    )
    return records



import swapanything
from packaging import version


def test_version() -> None:
    """Test that package has version, is semver and >= '0.0.1.dev0'"""
    # Test that package has readable version
    assert isinstance(swapanything.__version__, str)
    # Test that package has compliant semver
    assert version.Version(swapanything.__version__)
    # Test that package has semver greater or equal than '0.0.1.dev0'
    assert version.Version(swapanything.__version__) >= version.Version("0.0.1.dev0")



from packaging.version import Version
from pydantic import __version__

from ._base import BackendType

if Version(__version__) < Version("2.0"):  # pragma: no cover
    raise Exception("Swapanything Backend requires pydantic>=2.0")



from typing import Iterable, Optional

import networkx as nx
import numpy as np
import pandas as pd

"""
Select matches module. In this module you'll find everything related to the selection
of matches between individuals. This means you'll obtain a subset of the found matches among
all the feasible ones.
"""

def select_matches(
    matches: pd.DataFrame,
    subjects_col: str,
    slots_col: str,
    match_scores: Optional[pd.Series] = None,
    maxcardinality: Optional[bool] = None,
    return_graph: bool = False,
) -> pd.DataFrame:
    """
    Parameters
    ----------
    matches: pandas.DataFrame
             Dataframe containing the matches to be selected. The df is expected to have a column containing the time slots and
             a column couples of subjects. Each time slot has a unique couple assigned to it.
    subjects_col: str
                  Name of the column where subjects, namely individuals, are stored. The subjects are
                  expected to be unique.
    slots_col: str
               Name of the column where time slots indicating subjects availabilities are stored.
    match_scores: pandas.Series, default=None
                  Score defining the affinity of a match between individuals, higher score defines higher affinity.
                  Matching algorithm tries to maximize the weight of the selected edges.
    maxcardinality: bool, default=None
                    If maxcardinality is True, compute the maximum-cardinality matching with maximum
                    weight among all maximum-cardinality matchings.
    return_graph: bool, default=False
                  If True the function returns the graph obtained from the matches between individuals other than the
                  selected matches.

    Returns
    -------
    selected: pandas.DataFrame
    G: networkx.Graph, optional

    Raises
    ------
    AssertionError
        If the subjects in the matches table are not unique

    Examples
    --------
    all_possible_matches = be.get_all_matches()
    |   |                 subject |   availability |
    ------------------------------------------------
    | 0 |   (Barbarianna, Katana) |       (12:00,) |
    | 1 |   (Hackerman, KungFury) |       (10:00,) |
    | 2 | (Hackerman, Triceracop) |       (11:00,) |
    | 3 |      (Hoff 9000, T-Rex) |       (16:00,) |
    | 4 |        (KungFury, Thor) | (13:00, 14:00) |
    | 5 |  (KungFury, Triceracop) |        (9:00,) |
    | 6 |           (T-Rex, Thor) |       (15:00,) |

    select_matches(all_possible_matches, backend=be)
    |   |                 subject |   availability |
    ------------------------------------------------
    | 0 |   (Barbarianna, Katana) |       (12:00,) |
    | 1 | (Hackerman, Triceracop) |       (11:00,) |
    | 2 |      (Hoff 9000, T-Rex) |       (16:00,) |
    | 3 |        (KungFury, Thor) | (13:00, 14:00) |
    """
    assert matches[subjects_col].is_unique

    # Considering the table entry 9:00 (KungFury, Triceracop), we expand the subject couple to obtain the following dataframe:
    # _matches: _s1     | _s2
    #           KungFury| Triceracop
    _matches = matches[subjects_col].apply(pd.Series)
    _matches.columns = ("_s1", "_s2")
    _matches[slots_col] = matches[slots_col]
    if isinstance(match_scores, Iterable):
        _matches["_score"] = np.array(match_scores)
        maxcardinality = maxcardinality or False
    else:
        _matches["_score"] = 1
        maxcardinality = True

    G: nx.Graph = nx.from_pandas_edgelist(
        df=_matches, source="_s1", target="_s2", edge_attr=["_score", slots_col]
        )
    results_weighted: set = nx.algorithms.matching.max_weight_matching(
        G, maxcardinality=maxcardinality, weight="_score"
    )
    results_weighted = pd.Index(
        {tuple(sorted(x)) for x in results_weighted},
        tupleize_cols=False,
        name=subjects_col,
    )

    selected = (
        matches.set_index(subjects_col)
        .reindex(pd.Index(results_weighted, tupleize_cols=False))
        .sort_index()
        .reset_index()
    )

    if return_graph:
        return selected, G
    else:
        return selected



from typing import List

import networkx as nx
import pandas as pd
import pytest
from swapanything import select


@pytest.fixture
def subject_features() -> List[str]:
    return ["a", "b", "c"]


@pytest.fixture
def subjects_col() -> str:
    return "subj"


@pytest.fixture
def slots_col() -> str:
    return "avail"


@pytest.fixture
def exclusions_subject_columns() -> List[str]:
    return ["es1", "es2"]


@pytest.fixture
def possible_matches() -> List[List[tuple]]:
    possible_matches = [
        [("sub1", "sub2"), ("A",)],
        [("sub1", "sub3"), ("B",)],
        [("sub2", "sub3"), ("C",)],
        [("sub4", "sub5"), ("D",)],
        [("sub1", "sub6"), ("E", "F")],
        [("sub6", "sub7"), ("G",)],
        [("sub7", "sub8"), ("H",)],
    ]
    return possible_matches


@pytest.mark.parametrize(
    "scores,expected_result_ixs",
    [
        (
            None,
            [4, 2, 3, 6],
            # [("sub1", "sub6"), ("E", "F")],
            # [("sub2", "sub3"), ("C",)],
            # [("sub4", "sub5"), ("D",)],
            # [("sub7", "sub8"), ("H",)],
        ),
        (
            # to test mismatching index
            pd.Series([1.0, 1.0, 1.0, 1.0, 1.0, 9001.0, 1.0]),
            [0, 3, 5]
            # [("sub1", "sub2"), ("A",)],
            # [("sub4", "sub5"), ("D",)],
            # [("sub6", "sub7"), ("G",)],
        ),
        ([1.0, 1.0, 1.0, 1.0, 1.0, 9001.0, 1.0], [0, 3, 5]),
    ],
)
def test_select_matches(
    scores,
    expected_result_ixs,
    subjects_col: str,
    slots_col: str,
    possible_matches: List[List[tuple]],
) -> None:
    matches = pd.DataFrame(
        possible_matches,
        columns=[
            subjects_col,
            slots_col,
        ],
        index=[0] * 7,
    )

    results = select.select_matches(
        matches,
        subjects_col=subjects_col,
        slots_col=slots_col,
        match_scores=scores,
    )

    expected_result = matches.iloc[expected_result_ixs].reset_index(drop=True)
    assert isinstance(results, pd.DataFrame)
    assert results.columns[0] == subjects_col
    assert results.columns[1] == slots_col
    assert results.equals(expected_result)

    _, G = select.select_matches(
        matches,
        subjects_col=subjects_col,
        slots_col=slots_col,
        match_scores=scores,
        return_graph=True,
    )
    assert isinstance(G, nx.Graph)



from random import randint, seed
from typing import Tuple

import pandas as pd
import pytest
from faker import Faker
from faker.providers import DynamicProvider
from swapanything.prep import get_all_matches
from swapanything.select import select_matches

n = 42
samples = 500

seed(n)
index = [f"Sub{i+1}" for i in range(samples)]
# Create faker object
industries_provider = DynamicProvider(
    provider_name="industry",
    elements=[
        "AI/ML",
        "AR/VR",
        "Agritech",
        "Altro",
        "Blockchain",
        "Clean Energy",
        "Cybersecurity",
        "Data Analytics",
        "E-Commerce/Online Marketplaces",
        "Edtech",
        "Fintech",
        "Foodtech",
        "Gaming",
        "Greentech",
        "Healthtech/Medtech",
        "IoT",
        "ML",
        "Media & Entertainment",
        "Mobile Apps",
        "No Code/ Low Code",
        "Online Marketplaces",
        "SaaS",
        "Social Media",
        "Sportstech",
        "Wearable tech",
        "Web Development",
        "Web3",
    ],
)

roles_provider = DynamicProvider(
    provider_name="role",
    elements=[
        "CEO",
        "CTO",
        "COO",
        "CFO",
        "Marketing Manager",
        "Sales Manager",
        "Business Development Manager",
        "Product Manager",
        "Human Resources Manager",
        "Software Developers/Engineers",
        "UX/UI Designers",
        "Customer Support Manager",
        "Data Analyst",
        "Digital Content Specialist",
        "Public Relations Specialist",
    ],
)

growth_provider = DynamicProvider(
    provider_name="growth",
    elements=["Pre-Seed", "Seed", "Bootstrap", "Series A", "Verso l'infinito ed oltre"],
)

language_provider = DynamicProvider(
    provider_name="language",
    elements=["Italian", "English", "Spanish", "Frech", "German"],
)

availabilities_provider = DynamicProvider(
    provider_name="availability",
    elements=[
        "Mon 10:00",
        "Mon 12:00",
        "Mon 16:00",
        "Mon 18:00",
        "Thu 10:00",
        "Thu 12:00",
        "Thu 16:00",
        "Thu 18:00",
        "Fri 10:00",
        "Fri 12:00",
        "Fri 16:00",
        "Fri 18:00",
    ],
)

goals_provider = DynamicProvider(
    provider_name="goals",
    elements=[
        "Creare partnership e sinergie con altri Founders",
        "Fare del buon sano networking",
        "Trovare supporto su temi specifici",
        "Confrontarmi con founders nel mio stesso stadio di crescita",
        "Confrontarmi con founders più avanti di me",
        "Dare supporto a founders più indietro di me",
        "Trovare un Co-Founder",
        "Altro",
    ],
)

index_provider = DynamicProvider(
    provider_name="index",
    elements=index,
)


def fake_ind(min, max):
    return [fake.industry() for _ in range(randint(min, max))]


def fake_lang(min, max):
    return [fake.language() for _ in range(randint(min, max))]


def fake_goal(min, max):
    return [fake.goals() for _ in range(randint(min, max))]


def fake_avail(min, max):
    return [fake.availability() for _ in range(randint(min, max))]


fake = Faker("it_IT")
Faker.seed(n)
fake.add_provider(industries_provider)
fake.add_provider(roles_provider)
fake.add_provider(availabilities_provider)
fake.add_provider(index_provider)
fake.add_provider(growth_provider)
fake.add_provider(language_provider)
fake.add_provider(goals_provider)


@pytest.fixture(scope="session")
def get_large_datasets() -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    # Creation syntetic data
    name = [fake.unique.name() for _ in range(samples)]

    subjects_data = {
        "Index": index,
        "Nome e Cognome": name,
        "Industry": [
            fake_ind(0, 3) for _ in range(samples)
        ],  # Select randomly from zero to max 3 industries among the ones currently in the airtable database
        "Role": [fake.role() for _ in range(samples)],
        "Growth Stage": [fake.growth() for _ in range(samples)],
        "Languages": [
            fake_lang(1, 3) for _ in range(samples)
        ],  # Select randomly from 1 to max 3 languages among ['Italian', 'English', 'Spanish', 'Frech', "German"]
        "Goal": [fake_goal(1, 8) for _ in range(samples)],
    }  # Select randomly from 1 to max 8 goals among the ones currently in the airtable database

    availabilities_data = {
        "Index": index,
        "Nome e Cognome": name,
        "Availabilities": [fake_avail(1, 5) for _ in range(samples)],
    }  # Select randomly from 1 to max 5 slot availabilities among ["Mon 10:00", "Mon 12:00", "Mon 16:00", "Mon 18:00", "Thu 10:00", "Thu 12:00", "Thu 16:00", "Thu 18:00", "Fri 10:00", "Fri 12:00", "Fri 16:00", "Fri 18:00"]

    exclusions_data = {
        "SubjectA": [fake.unique.index() for _ in range(3)] + index,
        "SubjectB": [fake.unique.index() for _ in range(3)] + index,
    }

    # DataFrame Creation
    subjects = pd.DataFrame(subjects_data)
    availabilities = pd.DataFrame(availabilities_data).explode("Availabilities")
    exclusions = pd.DataFrame(exclusions_data)

    for i in range(len(exclusions)):
        if int(exclusions["SubjectA"][i][3:]) > int(exclusions["SubjectB"][i][3:]):
            temp = exclusions["SubjectA"][i]
            exclusions["SubjectA"][i] = exclusions["SubjectB"][i]
            exclusions["SubjectB"][i] = temp
        else:
            pass

    exclusions.drop_duplicates(inplace=True)
    return subjects, availabilities, exclusions


def test_large_dateset(
    get_large_datasets: Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame],
):
    _, availabilities, exclusions = get_large_datasets

    slots_col = "Availabilities"
    subjects_col = "Index"

    all_possible_matches = get_all_matches(
        availabilities=availabilities,
        subject_col=subjects_col,
        slot_col=slots_col,
    )
    assert isinstance(all_possible_matches, pd.DataFrame)
    assert all_possible_matches["Index"].is_unique

    selected = select_matches(
        all_possible_matches,
        subjects_col=subjects_col,
        slots_col=slots_col,
    )
    assert isinstance(selected, pd.DataFrame)
    assert selected["Index"].is_unique
    assert len(selected) < len(all_possible_matches)
